#!/usr/bin/perl -w
#
#######################################################
# LogDog
# Written by: Brandon Zehm <caspian@linuxfreak.com>
#
# License:
# LogDog (hereafter referred to as "program") is free software;
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation; either version
#   2 of the License, or (at your option) any later version.
# Note that when redistributing modified versions of this source code, you
#   must ensure that this disclaimer and the above coder's names are included
#   VERBATIM in the modified code.
#
# Disclaimer:
# This program is provided with no warranty of any kind, either expressed or
#   implied.  It is the responsibility of the user (you) to fully research and
#   comprehend the usage of this program.  As with any tool, it can be misused,
#   either intentionally (you're a vandal) or unintentionally (you're a moron).
#   THE AUTHOR(S) IS(ARE) NOT RESPONSIBLE FOR ANYTHING YOU DO WITH THIS PROGRAM
#   or anything that happens because of your use (or misuse) of this program,
#   including but not limited to anything you, your lawyers, or anyone else
#   can dream up.  And now, a relevant quote directly from the GPL:
#
#                           NO WARRANTY
#
#  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
# FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
# OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
# PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
# OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
# TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
# PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
# REPAIR OR CORRECTION.
#
#   ---
#
# Whee, that was fun, wasn't it?  Now let's all get together and think happy
#   thoughts.  And remember, the same class of people who made all the above
#   legal spaghetti necessary are the same ones who are stripping your rights
#   away with DVD CCA, DMCA, stupid software patents, and retarded legal
#   challenges to everything we've come to hold dear about our internet.
#   So enjoy your dwindling "freedom" while you can, because 1984 is coming
#   sooner than you think.  :[
#
#######################################################
#
# Purpose: To create a program that is capable of monitoring /var/log/messages
#          and /var/log/syslog and will email a list of administrators if 
#          adnormalities are found. 
#
# Layout:  Checks logfiles every $interval seconds, and emails the people in 
#          @admin_list if adnormalities are found.  If its a weekend or after
#          normal work hours it also executes $after_hours_command.
#

use Time::Local;

## Global Vars

$version   = 'v0.94.6';            ## Version
$conf      = '/etc/logdog.conf';   ## Configuration file location/name.
$0         = '[logdog.pl]';        ## Process display name
@critical_data = ();

## End Global Vars



## Catch all SIG's and exit cleanly
$SIG{'QUIT'} = \&exit;
$SIG{'TERM'} = \&exit;
$SIG{'INT'}  = \&exit;
$SIG{'KILL'} = \&exit;
$SIG{'HUP'}  = \&read_conf;  ## If HUP received, reload the configuration and keep running


## Check command line parameters.
$numargv = @ARGV;
$daemon = 'no';
for ($counter = 0; $counter < $numargv; $counter++) {
  if ($ARGV[$counter] eq '-c') {
    $counter++;
    $conf = $ARGV[$counter];
  }
  elsif ($ARGV[$counter] eq '--config') {
    my @temp = split (/=/, $ARGV[$counter]);
    $conf = $temp[1];
  }
  elsif (($ARGV[$counter] eq '-d') or ($ARGV[$counter] eq '--daemon')) {
    $daemon = 'yes';
  }
  elsif (($ARGV[$counter] eq '-v') or ($ARGV[$counter] eq '--version')) {
    printf "LogDog $version\n";
    exit 0;
  }
  elsif (($ARGV[$counter] eq '-h') or ($ARGV[$counter] eq '--help')) {
    printf "\n";
    printf "LogDog $version\n";
    printf "Written by: Brandon Zehm <caspian\@linuxfreak.com>\n";
    printf "\n";
    printf "Useage:  logdog.pl [options]\n";
    printf "\n";
    printf "         -d,--daemon        become a daemon\n";
    printf "         -c,--config=file   specify a different config file\n";
    printf "         -v,--version       show version informaion\n";
    printf "         -h,--help          display this screen\n";
    printf "\n";
    exit 0;
  }
}


## If the daemon flag was set then start running.
if ($daemon eq 'yes') {
  ## Make sure LogDog is not already running..
  my @ps = `ps ax`;
  @ps = (grep /\[logdog.pl\]/, @ps);
  my $run = @ps;
  if ($run > 1) {
    &read_conf;
    print "LogDog is already runnning.  Exiting.\n";
    &syslog("Another copy of LogDog is already runnning.  Exiting.");
    exit;
  }
  undef $daemon; 
  undef $numargv; 
  undef $counter;
  &daemon;
  exit 0;
}
# Otherwise something went wrong, so give a short help.
printf "Useage:  logdog.pl [-d,-c,-v,-h]\n";
exit 0;





### DAEMON LOOP (normal mode) ###
sub daemon {
  &read_conf;                                                      ## Read configuration file
  unless (fork) {                                                  ## Fork the process (freeing the terminal window)
    &syslog("LogDog started successfully");
    
    while (1) {                                                    #### START MAIN WHILE LOOP ###################################
      my $date = (time() - $interval);                             ## Get the current time and take off $interval seconds
      &parse_files($date) or &syslog("Error parsing files.");      ## Parse all file entries with a time of $date or higher
      &email_important_data or &syslog("Error emailing data.");    ## Email critical information
      sleep $interval;                                             ## Sleep
    }                                                              #### END MAIN WHILE LOOP #####################################
    exit 0;
  }
  exit 0;
}




#########################################################
# SUB: &read_conf
#      Reads and executes the configuration file and
#      warns if everything does not get set.
#########################################################

sub read_conf {
  ## Read and import configuration file
  do "$conf";
  
  if (!(defined $interval)) { 
    $interval = 30;
    &syslog("WARNING:  interval was not initialialized in $conf, using default value of $interval");
  }
  if (!(defined $email_server)) { 
    $email_server = 'localhost';
    &syslog("WARNING:  email_server was not initialialized in $conf, using default value of $email_server");
  }
  if (!(defined $email_from)) { 
    $email_from = 'logdog@localhost.com';
    &syslog("WARNING:  email_from was not initialialized in $conf, using default value of $email_from");
  }
  if (!(defined $email_program)) { 
    $email_program = '/usr/local/bin/sendEmail';
    &syslog("WARNING:  email_program was not initialialized in $conf, using default value of $email_program");
  }
  if ($admin_list[0] eq '') { 
    &syslog("WARNING:  admin_list was not initialialized in $conf, this is a critical field.  EXITING.");
    print "WARNING:  \@admin_list was not initialialized in $conf\n";
    print "          If this list is not set NO ONE will be notified\n";
    print "          when problems are found!\n";
    print "EXITING\n";
    exit 1;
  }
  if ($file_list[0] eq '') { 
    @file_list = (
                   '/var/log/messages',
                   '/var/log/syslog',
                 );
    &syslog("WARNING:  file_list was not initialialized in $conf, using default value.");
  }
  ## Verify that the files are valid and readable by this process
  else {
    foreach(@file_list) {
      if (!-e $_) {
        print "ERROR: File [$_] does not exist!  Hint: Check logdog.conf\n";
        exit;
      }
      if (!-r $_) {
        print "ERROR: File [$_] is not readable by me!\nLogDog needs read access to this file.\n";
        exit;
      }
    }
  }
  if ($warn_list[0] eq '') { 
    @warn_list = (
                   'fail',
                   'fatal',
                   'refuse',
                   'breakin',
                   'warning',
                   'error',
                   'unauthorized',
                   'detected',
                   'expired',
                   'useradd',
                   'userdel',
                   'passwd',
                   'bogus',
                   ' su\[', # ]
                 );
    &syslog("WARNING:  warn_list was not initialialized in $conf, using default value.");
  }
  if ($ignore_list[0] eq '') { 
    @ignore_list = ();
  }
}





#########################################################
# SUB: &parse_files
# Parses the log files and stores critical information
# in the global var @critical_data
# Returns 1 on success 0 on failure.
#########################################################
sub parse_files {
  my $cur_date = $_[0];
  my @buffer = ();                ## Gets used as a temp buffer
  my $num_of_files = @file_list;
  my @file = ();
  my $x = 0;
  my $file_number = 0;
  
  ## FOR EACH FILE:
  ## --------------
  for ($file_number = 1; $file_number <= $num_of_files; $file_number++) {
    open (FILE, "$file_list[$file_number - 1]")               ## Open the current file  
    || die("File [$file_list[$file_number - 1]] could not be opened for reading.\nHint: check logdog.conf\n\n");  ## Or die
    @file = readline *FILE;                                   ## and read it into @file
    close FILE;                                               ## and close the file.
    
    ## EXTRACT NEW LINES FROM @file
    ## ----------------------------
    my $cur_line = (scalar(@file) - 1);     ## $cur_line == the last line in the file
    
    my $new_material = 1;                   ## Assume we have new material in the log.
    my $line = "";                          ## Temp var for processing lines from the file.
    
    while (($new_material) && ($cur_line >= 0)) {
      $line = $file[$cur_line];
      
      ## Make sure the line looks like a real syslogged line
      if ($line =~ /:/) {
        my @temp1 = split (/ + ?/, $line);
        my @temp2 = split (/:/, $temp1[2]);
        
        $tmp = @temp2;
        if ($tmp <= 2) {                      ## If the line is not formatted correctly ignore it.
          $cur_line = ($cur_line - 1);        ## Go up one line in the file.
        }
        else {
          ## Basically the next 5 lines are to get the number of seconds since 1970
          ## out of the 'Feb 09 13:43:21' from the logfile line. 
          my $sec = $temp2[2];  my $min = $temp2[1];
          my $hour = $temp2[0]; my $mday = $temp1[1];
          my $mon = &return_month("$temp1[0]");
          my $year = `date \"\+\%Y\"`; chop $year;
          my $time = timelocal($sec, $min, $hour, $mday, $mon, $year);
          
          ## The entry time of $line is now compared with the current time to see if
          ## it was entered in the last $interval seconds.
          if (($time >= $cur_date) && ($time <= ($cur_date + $interval))) {
            push @critical_data, $line;                ## Add the curent line to @critical_data and
            $cur_line = ($cur_line - 1);               ## go up one line in the file.
          }
          ## If the line was too old, then exit this file, and go on to the next file.
          else {
            $new_material = 0;
          }
        }
      }
    }
  }
  
  ## REMOVE THE UNINPORTANT ENTRIES FROM @critical_data
  ## --------------------------------------------------
  my $important_line = "no";
  
  for ($cur_line = @critical_data; $cur_line > 0; $cur_line = ($cur_line - 1)) {   # For each line
    for ($tmp = @warn_list; $tmp > 0; $tmp = ($tmp - 1)) {                   ## Set the line to 'important' status
      if ($critical_data[$cur_line - 1] =~ /$warn_list[$tmp - 1]/i) {        ## if it matches one of the 'critical' words.
        $important_line = "yes";
      }
    }
    for ($tmp = @ignore_list; $tmp > 0; $tmp = ($tmp - 1)) {                 ## Set the line to 'not important' status
      if ($critical_data[$cur_line - 1] =~ /$ignore_list[$tmp - 1]/i) {      ## if it matches one of the 'ignore' words.
        $important_line = "no";
      }
    }
    if ($important_line =~ /yes/) {                        ## If the line is still important
      $critical_data[$cur_line - 1] =~ s/"/\\"/g;          ## Escape control chars
      $critical_data[$cur_line - 1] =~ s/\`/\\\`/g;
      $critical_data[$cur_line - 1] =~ s/\$/\\\$/g;
      push @buffer, $critical_data[$cur_line - 1];         ## put it in the outgoing list.
    }
  }
  
  @critical_data = @buffer;                                ## Put the outgoing buffer into @critical_data
  
  ## DONE!
  
  return(1);
}



#########################################################
# SUB: &email_important_data;
#########################################################
sub email_important_data {
  if (scalar(@critical_data)) { ## If there are entries in the critical_data
    
    ## Send email to everyone in the @admin_list
    system ("$email_program -s $email_server -t @admin_list -f $email_from -u \"LogDog: Critical data found in logs.\" -m \"@critical_data\" > /dev/null &");
    
    ## If $after_hours_command is defined and its after work hours or a weekend then execute it.
    if (length($after_hours_command) > 3) {
      my $hour = `date \+\%H`; chop $hour;
      my $day = `date \+\%w`; chop $day;
      
      ## Replace CRITICAL_DATA with the real critical data.
      my $tmp = $after_hours_command;
      $tmp =~ s/CRITICAL_DATA/\"@critical_data\"/g;
      if (($hour <= 7) or ($hour >= 18)) {
        system("$tmp > /dev/null &");
      }
      elsif (($day == 0) or ($day == 6)){
        system("$tmp > /dev/null &");
      }
    }
  }
  ## Clear @critical_data since we just sent it all off.
  @critical_data = ();
  ## Return success.
  return(1);
}


#########################################################
# SUB: &return_month(jan,feb,etc)
#  returns a number between 0 and 11 of the month.
#  returns 0 on error. (should be fixed)
#########################################################
sub return_month {
  my $x = $_[0];
  if ($x =~ /jan/i) { return  0; }
  if ($x =~ /feb/i) { return  1; }
  if ($x =~ /mar/i) { return  2; }
  if ($x =~ /apr/i) { return  3; }
  if ($x =~ /may/i) { return  4; }
  if ($x =~ /jun/i) { return  5; }
  if ($x =~ /jul/i) { return  6; }
  if ($x =~ /aug/i) { return  7; }
  if ($x =~ /sep/i) { return  8; }
  if ($x =~ /oct/i) { return  9; }
  if ($x =~ /nov/i) { return 10; }
  if ($x =~ /dec/i) { return 11; }
  return (0);
}



#########################################################
# SUB: &syslog($message)
#  Puts $message into $file_list[0] 
#  (usually /var/log/messages)
#########################################################
sub syslog {
  ## If the file is writeable to me, then write an entry.
  if (-w $file_list[0]) {
    my $message = $_[0];
    my $date = `date +\"%b %e %T\"`; chop $date;
    my $hostname = $ENV{'HOSTNAME'};
    $hostname =~ s/\..*$//;
    `echo "$date $hostname LogDog\[$$\]: $message" >> $file_list[0]`;
  }
}



#########################################################
# SUB: &exit
#  Exit's LogDog nicely
#########################################################
sub exit {
  $signame = shift;
  &syslog("Exiting normally, received SIG_$signame");
  exit 0;
}
